#include <string>
#include <sstream>
#include <cstdlib>
#include "berstream.h"

/* tda BER_Error */

char *BER_Error::messages[] = {
  "no error",
  "not enough memory",
  "out of range",
  "bad or incompatible iterator",
  "data in buffer and input operator do not match",
  "no sequence to end",
  "unknown error",
  0 };

BER_Error::BER_Error(Code _id) throw() : id(_id) { }
BER_Error::~BER_Error() throw() { }
const char *BER_Error::what() const throw() {
  return id > 0 && id < ERROR_COUNT ? messages[id] : messages[UNKNOWN];
}

/* tda char_buffer */

char_buffer::char_buffer(int _chunk_size) throw() :
  data(0),
  data_ptr(0),
  buffer_size(0),
  chunk_size(_chunk_size) { }

void char_buffer::copy(const char_buffer &cb) throw(BER_Error)
{
  data_ptr = cb.data_ptr;
  buffer_size = cb.buffer_size;
  chunk_size = cb.chunk_size;
  if ((data = (unsigned char *) malloc(buffer_size)) == 0)
    throw BER_Error(BER_Error::NOMEM);
  memcpy(data, cb.data, buffer_size);
}
  
char_buffer::char_buffer(const char_buffer &cb) throw(BER_Error) {
  copy(cb);
}
  
char_buffer& char_buffer::operator=(const char_buffer &cb) throw(BER_Error) {
  if (data != 0) free(data);
  copy(cb);
  return *this;
}

void char_buffer::reallocate() throw(BER_Error) {
  if ((data = (unsigned char *) realloc(data, buffer_size += chunk_size)) == 0)
    throw BER_Error(BER_Error::NOMEM);
}

void char_buffer::put(unsigned char c) throw(BER_Error) {
  if (data_ptr == buffer_size) reallocate();
  data[data_ptr++] = c;
}

char_buffer::~char_buffer() throw() {
  if (data != 0) free(data);
}

char_buffer::iterator char_buffer::begin() throw(BER_Error) {
  return iterator(data, &data[data_ptr], data);
}

char_buffer::iterator char_buffer::end() throw(BER_Error) {
  return iterator(data, &data[data_ptr], &data[data_ptr]);
}

char_buffer::iterator::iterator() throw() :
  data_begin(0), data_end(0), position(0) { }
char_buffer::iterator::iterator(unsigned char *_data_begin,
  unsigned char *_data_end, unsigned char *_position) throw(BER_Error):
  data_begin(_data_begin), data_end(_data_end), position(_position)
{
  if (position < data_begin || position > data_end)
    throw BER_Error(BER_Error::RANGE);
}
char_buffer::iterator::~iterator() throw() { }

char_buffer::iterator& char_buffer::iterator::operator=(const iterator &it)
  throw()
{
  data_begin = it.data_begin;
  data_end = it.data_end;
  position = it.position;
  return *this;
}

bool char_buffer::iterator::operator==(const iterator &it) const throw() {
  return data_begin == it.data_begin && data_end == it.data_end &&
    position == it.position;
}
bool char_buffer::iterator::operator<(const iterator &it) const
  throw(BER_Error) {
  if (data_begin != it.data_begin || data_end != it.data_end)
    throw BER_Error(BER_Error::BADITERATOR);
  return position < it.position; 
}
const char_buffer::iterator& char_buffer::iterator::operator++()
  throw(BER_Error) {
  if (position >= data_end) throw BER_Error(BER_Error::RANGE);
  position++;
  return *this;
}
const char_buffer::iterator char_buffer::iterator::operator++(int)
  throw(BER_Error) {
  iterator before = *this;
  ++*this;
  return before;
}

const char_buffer::iterator& char_buffer::iterator::operator--()
  throw(BER_Error) {
  if (position <= data_begin) throw BER_Error(BER_Error::RANGE);
  position--;
  return *this;
}
const char_buffer::iterator char_buffer::iterator::operator--(int)
  throw(BER_Error) {
  iterator before = *this;
  --*this;
  return before;
}

unsigned char char_buffer::iterator::operator*() const throw(BER_Error) {
  if (position < data_begin || position > data_end)
    throw BER_Error(BER_Error::RANGE);
  return *position;
}

/* ablony T_buffer, buffer a buffer<char> jsou v hlavikovm souboru */

/* tda berstream */

void berstream::writelength(unsigned long int length) throw(BER_Error) {
  if (length < 127)
    b->put(length & 0xff); /* mal dlka */
  else {
    int bytes = 1;
    unsigned long int length2 = length;
    while (length2 >>= 8) bytes++; /* zjistme, na kolik bajt je dlka */
    b->put(bytes | 0x80); /* a zapeme to */
    for (; bytes >= 0; bytes--) /* nyn zapeme onu dlku po bajtech */
      b->put((length >> (bytes-1)*8) & 0xff);
  }
}

unsigned long int berstream::readlength(char_buffer::iterator &i)
  throw(BER_Error) {
  int bytes;
  if ((bytes = *(i++)) & 0x80) {
    bytes &= ~0x80;
    unsigned int length = 0;
    for (; bytes >= 0; bytes--) /* nateme dlku po bajtech */
      length |= *(i++) << (bytes-1)*8;
    return length;
  }
  else
    return bytes;
}

/* tda oberstream */

oberstream& oberstream::operator<<(bool value) throw(BER_Error) {
  b->put(BOOL);
  b->put(sizeof(bool));
  b->put(value ? 0xff : 0x00);
  return *this;
}

oberstream& oberstream::operator<<(int value) throw(BER_Error) {
  unsigned int tmp = value > 0 ? value :
    (((1UL << sizeof(int)*8-1) - 1) << 1) + 2 + value;
  return *this << tmp;  
}

oberstream& oberstream::operator<<(unsigned int value) throw(BER_Error) {
  b->put(INT);
  b->put(sizeof(unsigned int));
  /* obecn zpis celho sla vymi bajty naped (big-endian) */
  for (int i = sizeof(unsigned int) - 1; i >= 0; i--)
    b->put((value >> i*8) & 0xff);
  return *this;  
}

oberstream& oberstream::operator<<(long int value) throw(BER_Error) {
  unsigned long int tmp = value > 0 ? value : 0xffffffff + value + 1;
  return *this << tmp;
}

oberstream& oberstream::operator<<(unsigned long int value) throw(BER_Error) {
  b->put(INT);
  b->put(sizeof(unsigned long int));
  b->put((value >> 24) & 0xff);
  b->put((value >> 16) & 0xff);
  b->put((value >>  8) & 0xff);
  b->put(value         & 0xff);
  return *this;
}

oberstream& oberstream::operator<<(const char *str) throw(BER_Error) {
  b->put(STRING);
  writelength(strlen(str));
  for (const char *p = str; *p != '\0'; p++)
    b->put(*p);
  return *this;
}

oberstream& oberstream::operator<<(const std::string& str) throw(BER_Error) {
  return *this << str.c_str();
}

void oberstream::encode(char_buffer &out) throw(BER_Error) {
  for (char_buffer::iterator bi = b->begin(); bi != b->end(); ++bi)
    out.put(*bi);
}

void oberstream::begin() throw(BER_Error) {
  char_buffer *sequence;
  if ((sequence = new char_buffer()) == 0)
    throw BER_Error(BER_Error::NOMEM);
  bufstack.push(b); /* ulome star pole, a je to cokoliv */
  b = sequence; /* zavedeme nov odkaz na pole */
}

void oberstream::end() throw(BER_Error) {
  if (bufstack.empty()) throw BER_Error(BER_Error::NOSEQ);
  char_buffer *sequence = b; /* chceme nynj rozdlanou sekvenci */
  b = bufstack.top(); bufstack.pop(); /* obnovme star pole */
  /* a vechno v sekvenci okoprujeme do hlavnho pole */
  b->put(SEQ);
  writelength(sequence->get_data_size());
  for (char_buffer::iterator j = sequence->begin(); j != sequence->end(); j++)
    b->put(*j);
  delete sequence;
}

/* tda iberstream */

iberstream::iberstream(const char_buffer &in) throw(BER_Error) :
  berstream(in), i(b->begin()) { }

bool iberstream::check(char_buffer::iterator &i,
  Type t, int len) throw(BER_Error) {
  Type type = static_cast<Type>(*(i++));
  if (type == SEQ) { /* sekvence ns nezajmaj, jen jejich obsah */
    readlength(i);
    return check(i, t, len);
  }
  if (type != t) return false;
  if (len == -1) return true;
  int l;
  if ((l = readlength(i)) != len) {
    for (; l >= 0; l--) ++i;
    return false;
  }
  return true;
}

iberstream& iberstream::operator>>(bool &value) throw(BER_Error) {
  if (!check(i, BOOL, sizeof(bool))) throw BER_Error(BER_Error::BADOPERATOR);
  value = *(i++);
  return *this;
}

iberstream& iberstream::operator>>(unsigned int &value) throw(BER_Error) {
  if (!check(i, INT, sizeof(int))) throw BER_Error(BER_Error::BADOPERATOR);
  value = 0;
  for (int j = sizeof(int) - 1; j >= 0; j--)
    value |= *(i++) << j*8;
  return *this;
}

iberstream& iberstream::operator>>(int &value) throw(BER_Error) {
  unsigned int tmp;
  *this >> tmp;
  if (tmp & (1 << sizeof(value)*8-1)) /* zporn slo */
    value = tmp & ~(1 << sizeof(value)*8-1) - (1 << sizeof(value)*8-1);
  else
    value = tmp;
  return *this;
}

iberstream& iberstream::operator>>(unsigned long int &value) throw(BER_Error){
  if (!check(i, INT, sizeof(long))) throw BER_Error(BER_Error::BADOPERATOR);
  value =  *(i++) << 24;
  value += *(i++) << 16;
  value += *(i++) << 8;
  value += *(i++);
  return *this;
}

iberstream& iberstream::operator>>(long int &value) throw(BER_Error) {
  unsigned long int tmp;
  *this >> tmp;
  if (tmp & 0x80000000)
    value = (tmp & 0x7fffffff) - 0x80000000;
  else
    value = tmp;
  return *this;
}

iberstream& iberstream::operator>>(std::string &str) throw(BER_Error) {
  if (!check(i, STRING)) throw BER_Error(BER_Error::BADOPERATOR);
  unsigned long int length = readlength(i);
  str.clear();
  for (unsigned long int l = 0; l < length; l++) str.append(1, *(i++));
  return *this;
}
